if (typeof module != "undefined")
{
	module.exports = {
		parseFSstatAsJSON:parseFSstatAsJSON,
		parseANOjsonForStats:parseANOjsonForStats,
		createRectNiiFromPoly:createRectNiiFromPoly,
        rendersubRoiIntoRoi:rendersubRoiIntoRoi,
        createNiftiMaskBuffer:createNiftiMaskBuffer

	}
}

sform2quaternion = function(mat)
{
 
   var vox = [ (math.norm([mat._data[0][0],mat._data[1][0],mat._data[2][0] ])),
               (math.norm([mat._data[0][1],mat._data[1][1],mat._data[2][1] ])),
               (math.norm([mat._data[0][2],mat._data[1][2],mat._data[2][2] ])),1];
   var ivox = [1/vox[0],1/vox[1],1/vox[2],1]
   var R = math.inv(math.multiply(mat,math.diag(ivox)));

   var sign = Math.sign(math.det(R));

   R = math.multiply(math.diag([1, 1, sign, 1]),R);

   var Rxx = R._data[0][0];
   var Rxy = R._data[0][1];
   var Rxz = R._data[0][2];
   var Ryx = R._data[1][0];
   var Ryy = R._data[1][1];
   var Ryz = R._data[1][2];
   var Rzx = R._data[2][0];
   var Rzy = R._data[2][1];
   var Rzz = R._data[2][2];
   var trace = Rxx+Ryy+Rzz;


   var w = 0;
   if (trace+1 > 0)
      w = Math.sqrt( trace + 1 ) / 2;

   var x = Math.sqrt( 1 + Rxx - Ryy - Rzz ) / 2;
   var y = Math.sqrt( 1 + Ryy - Rxx - Rzz ) / 2;
   var z = Math.sqrt( 1 + Rzz - Ryy - Rxx ) / 2;


   if (w >= x && w >= y && w >= z)
   {
        x = ( Rzy - Ryz ) / (4*w);
        y = ( Rxz - Rzx ) / (4*w);
        z = ( Ryx - Rxy ) / (4*w);
   }

   if (x >= w && x >= y && x >= z)
   {
        w = ( Rzy - Ryz ) / (4*x);
        y = ( Rxy + Ryx ) / (4*x);
        z = ( Rzx + Rxz ) / (4*x);
   }

   if (y >= w && y >= x && y >= z)
   {
        w = ( Rxz - Rzx ) / (4*y);
        x = ( Rxy + Ryx ) / (4*y);
        z = ( Ryz + Rzy ) / (4*y);
   }

   if (z >= w && z >= x && z >= x)
   {
        w = ( Ryx - Rxy ) / (4*z);
        x = ( Rzx + Rxz ) / (4*z);
        y = ( Ryz + Rzy ) / (4*z);
   }
   var eps = 0.0000000001;

   var quatern_b = -x * Math.sign(w-eps);
   var quatern_c = -y * Math.sign(w-eps);
   var quatern_d = -z * Math.sign(w-eps);   

   var qquat = { q: [quatern_b,quatern_c,quatern_d],
            pixdim: [sign,vox[0],vox[1],vox[2]] };

/*
var q = qquat.q
var a = Math.sqrt(Math.max(0.0,1.0-(q[0]*q[0]+q[1]*q[1]+q[2]*q[2])));
var b = q[0];
var c = q[1];
var d = q[2];
var pixdim = qquat.pixdim;
var qfac = pixdim[0];
   var a = Math.sqrt(Math.max(0.0,1.0-(b*b+c*c+d*d)))
  var e = [
      [pixdim[1]*(a*a+b*b-c*c-d*d),pixdim[1]*(2*b*c+2*a*d),pixdim[1]*(2*b*d-2*a*c)],
      [pixdim[2]*(2*b*c-2*a*d),pixdim[2]*(a*a+c*c-b*b-d*d),pixdim[2]*(2*c*d+2*a*b)],
      [qfac*pixdim[3]*(2*b*d+2*a*c),qfac*pixdim[3]*(2*c*d-2*a*b),qfac*pixdim[3]*(a*a+d*d-c*c-b*b)]]

console.log(print_matrix(e));
*/


   return qquat;
}


function parseANOjsonForStats(buf)
{
    var json = JSON.parse(buf);
    var ano = json.annotations;
    var result = {};
    for (var k = 0;k < ano.length;k++)
    {
    	var x = ano[k];
    	if (x.type == 'electrode')
    	{
          var ps = x.points;
          var avg = [0,0,0];
          var cnt = 0;
          for (var j = 0;j < ps.length;j++)
          {
              var pt = ps[j].coords;
              var active = parseInt(ps[j].active);
              if (math.abs(active) > 0)
              {
                  for (var i = 0;i < 3;i++)
                      avg[i] += pt[i];
                  cnt++;
              }
          }
          var c = {x:avg[0]/cnt,y:avg[1]/cnt,z:avg[2]/cnt,cnt:cnt};
    	}
    	else
    	{
          var ps = x.points;
          var r = [];          
          for (var j = 0;j < ps.length;j++)
          {
              ps[j].px = ps[j].coords[0];
              ps[j].py = ps[j].coords[1];
              ps[j].pz = ps[j].coords[2];
              delete ps[j].coords
              r.push(ps[j])
          }
          var c = r;
    	}


    	result[x.name] = c;
    }
    return result;

}

function parseFSstatAsJSON(buf)
{
    var lines = buf.split("\n");
    var cols = {};
    var byname = {};
    var k;
    for (k = 0; k < lines.length;k++)    
    {
        var x = lines[k];
        var tok = x.split(" ").filter((x) => (x != ""))
        if (tok[0] != "#")
            break;
        if (tok[1] == 'TableCol')
        {
            var colnum = parseInt(tok[2])-1;
            if (cols[colnum] == undefined)            
                cols[colnum] = {colnum:colnum};
            var c = cols[colnum];
            c[tok[3]] = tok[4];

        }
    }
    for (var j in cols)
    {
        byname[cols[j]['ColHeader']] = cols[j];
    }

    var result = {};

    for (;k<lines.length;k++)
    {
        var toks = lines[k].split(/[\ ]+/).filter((x) => (x != ""))
        var structname = toks[byname['StructName'].colnum];
        var p = {}
        for (var j = 0;j < toks.length;j++)
        {
            var q = cols[j].ColHeader;
            p[q] = toks[j];
        }
        result[structname] = p;
    }
    return result;


}



function createRectNiiFromPoly(polygons,inplaneres)
{
	var poly = polygons[0].subpoints;
	var sums = [0,0,0]
	for(var k=0;k<poly.length; k++)
	{
		var p1 = poly[k];
		var p2 = poly[(k+1)%poly.length];
		var p3 = poly[(k+2)%poly.length];

		sums[0] += (p1[2]-p2[2])*(p2[1]-p3[1]) - (p1[1]-p2[1])*(p2[2]-p3[2]);
		sums[1] += (p1[0]-p2[0])*(p2[2]-p3[2]) - (p1[2]-p2[2])*(p2[0]-p3[0])
		sums[2] += (p1[1]-p2[1])*(p2[0]-p3[0]) - (p1[0]-p2[0])*(p2[1]-p3[1])
		break;
	}


    var norm = math.sqrt(sums[0]*sums[0]+sums[1]*sums[1]+sums[2]*sums[2])
    var nz = [sums[0]/norm,sums[1]/norm,sums[2]/norm]
    var nx = math.cross([1,0,0],nz,true)._data;

    if (isNaN(nx[0]))
       nx = math.cross([0,1,0],nz,true)._data;
    if (isNaN(nx[0]))
       nx = math.cross([0,0,1],nz,true)._data;



    var ny = math.cross(nx,nz,true)._data;

    
    var aligned =[];
    var mini = [Infinity,Infinity,Infinity]
    var maxi = [-Infinity,-Infinity,-Infinity]
    var mindist = Infinity;


	for (var j = 0; j < polygons.length;j++)
	{
    	var poly = polygons[j].subpoints;
    	var p = [];
    	aligned.push(p)
    	for (var k = 0; k < poly.length;k++)
    	{
    	  var c = [poly[k][0]*nx[0]+poly[k][1]*nx[1]+poly[k][2]*nx[2],
    	           poly[k][0]*ny[0]+poly[k][1]*ny[1]+poly[k][2]*ny[2],
    	           poly[k][0]*nz[0]+poly[k][1]*nz[1]+poly[k][2]*nz[2] ]
    	  p.push(c)
    	  for (var r = 0; r < 3;r++)
    	  {
    	      mini[r] = math.min(mini[r],c[r])
    	      maxi[r] = math.max(maxi[r],c[r])
		  }
    	}
    	if (j>0)
    	{
	    	var dif = math.abs(aligned[j][0][2]-aligned[j-1][0][2]);
	    	if (dif > 0.2)
    	    	mindist=math.min(mindist,dif)
    	}
	}




	if (inplaneres == undefined)
		inplaneres = [mindist*0.5,mindist*0.5]

    if (mindist == Infinity)
    {
       if (inplaneres[0] == Infinity)
       {
		  inplaneres = [1,1];
		  mindist = 1;
       }
       else
       	 mindist = inplaneres[0];
    }
    
    var vsz = [inplaneres[0],inplaneres[1],mindist]

	var maxshape = [200,200,200];
	for (var k = 0; k < 3;k++)
		{
			if ((maxi[k]-mini[k])/vsz[k] > maxshape[k])
				vsz[k] = (maxi[k]-mini[k])/maxshape[k];
		}

    
	for (var k = 0; k < 3;k++)
	{
		maxi[k] += vsz[k];
		mini[k] -= vsz[k];
	}

	var or = [ mini[0]*nx[0] + mini[1]*ny[0] + mini[2]*nz[0],
	           mini[0]*nx[1] + mini[1]*ny[1] + mini[2]*nz[1],
	           mini[0]*nx[2] + mini[1]*ny[2] + mini[2]*nz[2] ]
    var edges = [[vsz[0]*nx[0],vsz[1]*ny[0],vsz[2]*nz[0],or[0]], 
                 [vsz[0]*nx[1],vsz[1]*ny[1],vsz[2]*nz[1],or[1]], 
                 [vsz[0]*nx[2],vsz[1]*ny[2],vsz[2]*nz[2],or[2]], 
                 [0,0,0,1]];
    
    var shape = [];
    for (var k = 0; k < 3;k++)
       shape[k] = math.ceil((maxi[k]-mini[k])/vsz[k])+1;
    
	console.log("rendering mask from contour - grid vsz:" +vsz + "  grid shape:"+shape )
    
	var tmp = createNiftiMaskBuffer(math.matrix(edges),vsz,shape)

    var nii = { edges: math.matrix(edges),  bbox: undefined, voxSize: vsz, sizes: shape, pixdim: [3,vsz[0],vsz[1],vsz[2]] ,
                wid:shape[0],widhei:shape[1]*shape[0],permutationOrder:[0,1,2],
                data: tmp.data ,buffer:tmp.buf}


    for (var k =0; k< aligned.length;k++)
    {
        var p = aligned[k];
        var poly = []
        for (var j = 0 ; j < p.length;j++)
            poly.push( math.round((p[j][0]-mini[0])/vsz[0]) , math.round((p[j][1]-mini[1])/vsz[1]) );
        poly.push(poly[0],poly[1])
    	var idx = math.round((p[0][2]-mini[2])/vsz[2])*shape[0]*shape[1];
        var f = ContainsPointFast(poly);
		for (var y = 0; y < shape[1];y++)
            for (var x = 0; x < shape[0];x++)
            	if (f(x,y))
            	{
            		nii.data[x+shape[0]*y+idx] = 1;
            	}
    }

    return nii;
}



var ContainsPointFast = function(pp)
{

    // close polygon
    pp.push(pp[0]);
    pp.push(pp[1]);
    var npoints = pp.length >> 1;
    var j = pp.length-2 ;
    var constant = [];
    var multiple = [];
    for (var i = 0; i < npoints-1;i++)
    {
        constant.push(pp[2*i]-(pp[2*i+1]*pp[2*j])/(pp[2*j+1]-pp[2*i+1])+(pp[2*i+1]*pp[2*i])/(pp[2*j+1]-pp[2*i+1]))
        multiple.push((pp[2*j]-pp[2*i])/(pp[2*j+1]-pp[2*i+1]));
        j=i;
    }
    return function(x,y)
    {
      var oddNodes=false
      var current=pp[2*(npoints-1)+1]>y
      var previous;
      for (var i=0; i<npoints; i++) {
        previous=current; 
        current=pp[2*i+1]>y; 
        if (current!=previous) 
            oddNodes^=y*multiple[i]+constant[i]<x; 
      }
      return oddNodes; 
    }


} 


function rendersubRoiIntoRoi(nii,polymask)
{
	var e = math.multiply(math.inv(polymask.edges),nii.edges)._data;
	var s0 = polymask.sizes[0]            
	var s1 = polymask.sizes[1]
	var s2 = polymask.sizes[2]
	var changedPoints = [];
	for (var z = 0; z < nii.sizes[2];z++)
	  for (var y = 0; y < nii.sizes[1];y++)
		 for (var x = 0; x < nii.sizes[0];x++)
		 {
			 var i = math.round(e[0][0]*x + e[0][1]*y + e[0][2]*z + e[0][3])
			 var j = math.round(e[1][0]*x + e[1][1]*y + e[1][2]*z + e[1][3])
			 var k = math.round(e[2][0]*x + e[2][1]*y + e[2][2]*z + e[2][3])
			 if (i>=0 & j>=0 & k>=0 & i < s0 & j < s1 & k < s2 & polymask.data[i+j*polymask.wid + k *polymask.widhei]>0)
				 {
					var idx = x+y*nii.wid+z*nii.widhei
					nii.data[idx] = 1;
					changedPoints.push(idx)
				 }
		 }
    return changedPoints
}


 
function createNiftiMaskBuffer(edges,voxSize,sz)
{

        // this is a default nifti buffer in uint8 representation.
        var niihdr = [92,1,0,0,117,105,110,116,49,54,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,114,0,1,0,16,0,16,0,1,0,1,0,1,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4,0,16,0,0,0,0,0,128,191,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,128,63,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,110,111,110,101,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,0,0,0,128,63,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,191,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,191,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,63,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,110,105,49,0,0,0,0,0];  

        var vx,vy,vz, edges;
        var littleEndian = true;

        var buffer = (new Uint8Array(niihdr)).buffer;


        var sx =  sz[0]
        var sy =  sz[1]
        var sz =  sz[2]

        var datasize = sx*sy*sz;
        
        vx =  voxSize[0]
        vy =  voxSize[1]
        vz =  voxSize[2]


    // set the  header to nifti header
       var view = new DataView(buffer);


       var signum_edges = Math.sign(math.det(edges._data));

       var qquat = sform2quaternion(edges);

       view.setFloat32(256, qquat.q[0], littleEndian);
       view.setFloat32(260, qquat.q[1], littleEndian);
       view.setFloat32(264, qquat.q[2], littleEndian);
       view.setFloat32(268, edges._data[0][3], littleEndian);
       view.setFloat32(272, edges._data[1][3], littleEndian);
       view.setFloat32(276, edges._data[2][3], littleEndian);

       var edges_ = edges;
                        
       edges = edges._data;

       // nifti standard is row-wise storage, edges already come in this from from daikon.js
       edges = edges[0].concat(edges[1]).concat(edges[2]).concat(edges[3]);

       var ndims = 3
      
       // sizes
       view.setInt32(0, 348,  littleEndian )
       view.setInt16(40, ndims,  littleEndian )
       view.setInt16(42, sx,  littleEndian )
       view.setInt16(44, sy,  littleEndian )
       view.setInt16(46, sz,  littleEndian )
       view.setInt16(48, 1,  littleEndian )
       view.setInt16(50, 1,  littleEndian )

       // edges
       var srow = new Float32Array(12);
       for(var i=0; i<12; i++) 
            srow[i] = view.setFloat32(280+4*i, (edges[i]), littleEndian)


       // ====== apply some other important stuff
       // set the magic number to n+1
       view.setInt32(344, 1848324352,  !littleEndian )
       // set vox offset to 352
       view.setFloat32(108, 352, littleEndian) 


       // pixdims
       view.setFloat32(76, signum_edges,  littleEndian ) // this is the qfac. Only set, if qform is provided??
       view.setFloat32(80, vx,  littleEndian )
       view.setFloat32(84, vy,  littleEndian )
       view.setFloat32(88, vz,  littleEndian )

       // qform=0, 
       view.setInt16(252, 0,  littleEndian )

      // datatype uint8
        view.setInt16(70, 2, littleEndian) 


       // combine the header and the imageData
       var tmp = new Uint8Array(buffer.byteLength + datasize);
       tmp.set(new Uint8Array(buffer), 0);

       return {buf:tmp,data: new Uint8Array(tmp.buffer,buffer.byteLength)};

}

