
/// fiberQuery ///////////////////////////////////


function FiberQuery(atlas,tracts,index)
{

    var fiberQuery = {involved_Labels:{}};
 
/*
 {  
    tracts : set
        set of tract indices resulting from the query
    labels : set
        set of labels resulting by the query
    endpoint : (set, set)
        sets of labels of where the tract endpoints are
 }
*/

    
    fiberQuery.traversing = function(a)
    {
    	if (index[a] == undefined)
    	{
    		if (atlas.labels[a] != undefined)
    			 throw new Error("label " + atlas.labels[a].name + " undefined");
    		else
    			 throw new Error("label " + a + " undefined");
    	}
    	else
    	{
			var labels ={}; labels[a]=true;
			//fiberQuery.involved_Labels[a] = true;
			return {tracts: index[a].traversing, endpoint:index[a].endpoint_labels,  labels : labels };
    	}
    }

    fiberQuery.endpoints_in = function(a)
    {
        var tracts = {};
        var endpoint = {};
        var labels = {};

        for (k in a.labels)
        {
             labels[k] = true;
             var tracts_with_given_endpoints = index[k].endpoint_in;
             var labels_of_endpoints = index[k].endpoint_labels;
             for (var j in tracts_with_given_endpoints)
             {
                tracts[j] = true;
                endpoint[k] = true;

             }

             fiberQuery.involved_Labels[k] = true;
        }

        return {tracts:tracts, endpoint: endpoint,  labels : labels };
    }

    fiberQuery.both_endpoint_in = function(a)
    {

    }

    fiberQuery.only = function(a)
    {

    }

    fiberQuery.union = function(a,b)
    {
      if (b != undefined)
		  return { tracts : union(a.tracts,b.tracts), 
				   labels : union(a.labels,b.labels) };
	  else
		  return { tracts : union(a.tracts,{}), 
				   labels : union(a.labels,{}) };
    }

    fiberQuery.intersect = function(a,b)
    {
      return { tracts : intersect(a.tracts,b.tracts), 
               labels :     intersect(a.labels,b.labels) };
    }

    fiberQuery.diff = function(a,b)
    {
      return { tracts : diff(a.tracts,b.tracts), 
               labels :     diff(a.labels,b.labels) };
    }

	// all the spatial rels
    function direction_query(which,query)
    {
       var all_labels = []
     
       var max = [-10000,-10000,-10000];
       var max_idx = [-1,-1,-1]
       var min = [10000,10000,10000];
       var min_idx = [-1,-1,-1]
     
       for (var k in query.labels)
       {
       		for (var j = 1; j < 3;j++)
				if (index[k].bbox.max[j] > max[j])
				{
					max_idx[j] = k; 
					max[j] = index[k].bbox.max[j];
				}
       		for (var j = 1; j < 3;j++)
				if (index[k].bbox.min[j] < min[j])
				{
					min_idx[j] = k; 
					min[j] = index[k].bbox.min[j];
				}
            all_labels.push(index[k][which]);
       }
       var tracts = intersect_array(all_labels);
	   var res = {tracts:tracts,labels:{}};
	   if (which == "anterior_of")
	   		res.labels =  index[min_idx[1]].labels_anterior_of;
	   if (which == "posterior_of")
	   		res.labels =  index[max_idx[1]].labels_posterior_of;
	   if (which == "inferior_of")
	   		res.labels =  index[min_idx[2]].labels_inferior_of;
  		 if (which == "superior_of")
	   		res.labels =  index[min_idx[2]].labels_superior_of;


       return res;
    }

    var types = ["anterior_of", "posterior_of",  "superior_of",  "inferior_of",  "lateral_of", "medial_of" ];
    for (var k = 0;k < types.length; k++)
        fiberQuery[types[k]] = function(which) { return function(query) { return direction_query(which,query);} }(types[k]);




    return fiberQuery
}



/// creates the index ///////////////////////////////////////////////////

function createWMQLindex(atlas,tractography,deffield,progress)
{
	var index = {};

    var labels = atlas.labels;
    var atlas_data = atlas.data;

    var tracts = tractography.tracts;
    var bbox_max = tractography.tracts_max;
    var bbox_min = tractography.tracts_min;

    var e ;
    if (atlas.invedges)
    	e = atlas.invedges;
    else
    	e = math.inv(atlas.edges)._data;
    var sz = atlas.sizes;
		

	var tolerance = 5; //?

	progress("spatial indexing")

	/* the_index 
	{
	   traversing      : set of tracts travering label (key)
	   ...._of         : set of labels with spatial relationship
	   endpoint_in     : set of tracts that have the label at one of its terminals
	   endpoint_labels : set of labels that <traversing> tracts visit 
	}
	*/
	for (var key in labels)
	{

		key = parseInt(key);
		var the_index =  { traversing: {},  labels_anterior_of :{}, labels_posterior_of:{}, labels_superior_of:{},labels_inferior_of:{},labels_medial_of:{},labels_lateral_of:{},
							anterior_of :{}, posterior_of:{}, superior_of:{},inferior_of:{},medial_of:{},lateral_of:{},endpoint_in:{},endpoint_labels:{} };
		index[key] = the_index;
		var label = labels[key];
		var label_bbox = label.bbox;		
		if (label_bbox == undefined)
			continue;

		the_index.bbox = label_bbox;

		//// labels
		for (key2 in labels)		
		{
			if (labels[key2].bbox == undefined)
				continue;

			key2 = parseInt(key2);
			
			// posterior/anterior
			if (labels[key2].bbox.max[1] < label_bbox.min[1] + tolerance)
				the_index.labels_posterior_of[key2] = true;
			if (labels[key2].bbox.min[1] > label_bbox.max[1] - tolerance)
				the_index.labels_anterior_of[key2] = true;			
			// inferior/superior
			if (labels[key2].bbox.max[2] < label_bbox.min[2] + tolerance)
				the_index.labels_inferior_of[key2] = true;
			if (labels[key2].bbox.min[2] > label_bbox.max[2] - tolerance)
				the_index.labels_superior_of[key2] = true;			
		}







		//// tracts
		for (var k = 0;k < tracts.length;k++)
		{
			
			// medial/lateral
			if (label_bbox.max[0] < 0)
			{   

				if (bbox_min[3*k+0] + tolerance > label_bbox.max[0])
				   the_index.medial_of[k] = true;
		  	    if (bbox_max[3*k+0] < label_bbox.min[0] + tolerance)
				   the_index.lateral_of[k] = true;
			}
			else
			{
				if (bbox_min[3*k+0] + tolerance > label_bbox.max[0])
				   the_index.lateral_of[k] = true;
		  	    if (bbox_max[3*k+0] < label_bbox.min[0] + tolerance)
				   the_index.medial_of[k] = true;

			}
				 
			// posterior/anterior
			if (bbox_max[3*k+1] < label_bbox.min[1] + tolerance)
				the_index.posterior_of[k] = true;
			if (bbox_min[3*k+1] + tolerance > label_bbox.max[1])
				the_index.anterior_of[k] = true;

			// inferior/superior
			if (bbox_max[3*k+2] < label_bbox.min[2] + tolerance)
				the_index.inferior_of[k] = true;
			if (bbox_min[3*k+2] + tolerance > label_bbox.max[2])
				the_index.superior_of[k] = true;
		}
	}

	// world coordinate to label
	var w2key = function (q0,q1,q2)
	{
		var p = [e[0][0]*q0 + e[0][1]*q1 + e[0][2]*q2 + e[0][3],
				e[1][0]*q0 + e[1][1]*q1 + e[1][2]*q2 + e[1][3],
				e[2][0]*q0 + e[2][1]*q1 + e[2][2]*q2 + e[2][3]];
		return atlas_data[sz[0]*sz[1]*math.round(p[2]) + sz[0]*math.round(p[1]) + math.round(p[0])];
	}

	if (deffield)
	{
		w2key = KAtlasTool.updateGetPixelFun(atlas,{},undefined,undefined,deffield,true)
	}



	progress("label indexing")


	for (var k = 0;k < tracts.length;k++)
	{
		var tract = tracts[k];
		
	    var l = tract.length;

	    var key1 = w2key(tract[0],tract[1],tract[2]);
	    var key2 = w2key(tract[l-3],tract[l-2],tract[l-1])
	    if (index[key1])
	    	index[key1].endpoint_in[k] = true;
	    if (index[key2])
		    index[key2].endpoint_in[k] = true;


	    for (var j = 0; j < l/3;j++)
	    {
		  var key = w2key(tract[3*j],tract[3*j+1],tract[3*j+2]);
		  if (index[key])
		  {
		  	index[key].traversing[k] = true;
		  	index[key].endpoint_labels[key1] = true;
		  	index[key].endpoint_labels[key1] = true;
		  }
	    }

	}
	return {index:index};

}






function processSelection(selection)
{

   var todisplay;
 
   if (selection.symbols["ans"] != undefined)
     todisplay = "ans";
   else
     todisplay = Object.keys(selection.to_display)[0];

   var selfun = selection.symbols[todisplay];
   if (selfun != undefined)
   {
   		var res = selfun('left')
        return {tracts:Object.keys(res.tracts),labels:(res.labels)};
   }
   else 
        return "nothing to display";        
}




