{

  /// some simple set ops for testing
  
  function union(a,b)
  {
    	var r = {};
    	for (var k in a)
    		r[k] = true;
    	for (var k in b)
    		r[k] = true;
    	return r;
  }
  
  function intersect(A,B)
  {
 	 var C = {};
    for (var x in A)
    {
    if (B[x])
        C[x] = true;
    }
    return C;
  }
  
  function diff(A,B)
  {
  
 	 var C = {};
    for (var x in A)
    {
    if (B[x] == undefined)
        C[x] = true;
    }
    return C;
  }
  
  function isSubset(A,B)
  {
    for (var x in A)
    {
    if (B[x] == undefined)
        {
        return false;
        }
    }
    return true;  
  }
  
  function set(x)
  {
       var s ={}; s[x] = true; return s; 
  }
    
  function tracts_traversing_label(label)
  {
       return set(label);  
  }
    
  
  //////// Fiberquery, interface 
  
  	var symbol_table = { };
  
  	var symbol_todisplay = {};
    
    
    if (typeof FiberQueryCurrent != "undefined")
    {
     	function extend(a,b)
          {
            for (var x in b)
              a[x] =b[x];
            return a;
          }    	
	     var FiberQuery = FiberQueryCurrent;
	     if (FiberQuery.symbol_table)
	     	symbol_table = extend( symbol_table,FiberQuery.symbol_table);
    }
    else
    {
      var FiberQuery = function () {}

      FiberQuery.endpoint_in = function(a)
      {
          return "test" + a;
      }


      FiberQuery.traversing = function(a)
      {
          return set(a);
      }

      FiberQuery.union = function(a,b)
      {
          if (b==undefined)
             return union({},a)
          else
             return union(a,b);
      }

      FiberQuery.intersect = function(a,b)
      {
          return intersect(a,b);
      }

      FiberQuery.diff = function(a,b)
      {
          return diff(a,b);
      }

      FiberQuery.isSubset = function(a,b)
      {
          return isSubset(a,b);
      }
	}    
    
    // adds a symbol to the symbol_table as 
    // a function depending on lhs parameters (hemi -> left,side,...)
    function addsymbol(identifier,expression,hemi)
    {
    	if (hemi != null)
        { 
            var hemisphere = hemi[1];
            var old_symbol = symbol_table[identifier];
            if (old_symbol == null) // for lhs side with side/opposite there should be just one definition
            {
            	if (hemisphere == "opposite")
            		symbol_table[identifier] = invert(expression);
                else
            		symbol_table[identifier] = expression;
            }
            else                  // this could be left/right or nothing
        		symbol_table[identifier] = function(side) 
                		{ return (side == hemisphere)?
                        			expression(side):old_symbol(side); }            	
        }
        else
            symbol_table[identifier] =  expression        
    }
    
	function invert(fun)  // to flip left/right behavior for a given function
    {
        return function(side) {
                  if (side == 'left')
	                  return fun('right');
                  else if (side == 'right')
    	              return fun('left');
                  else if (side == 'side')
    	              return fun('opposite');
                  else
    	              return fun('side');
                  }
    }
                  
}

////////////////////////////////////////////////////////////////////////

WMQL_script = (WMQL_statement / COMMENT / NEWLINE )*  ("#" comtext)?
	{  
		  if (typeof FiberQueryCurrent != "undefined") // this production mode
		  {
			  return {
			    symbols:symbol_table, 
			    to_display:symbol_todisplay     
			  };
		  }
		  else 
		  {
			  return {
				  ans: symbol_table['ans'](),
				  to_display:symbol_todisplay    
			  };

		  }
    }

comtext = [^\n\r]*

COMMENT = ("#" comtext [\n\r]) 


WMQL_statement = _ (WMQL_query) [\ \t\n\r;]+


WMQL_query = value:(assignment / blocked_assignment / visexpression) 

NEWLINE = ["\u000C""\r""\n"]+

_ "whitespace" = [ \t\n\r]* 
w "forcewhitespace" = [ \t\n\r]+

hemisphere = "left" / "right" / "side" / "opposite" { return text()}


/// assignments /////////////////////////////////


visexpression = expression:expression {
    addsymbol("ans",expression,null);
    symbol_todisplay["ans"] = true;
}

assignment = identifier:identifier hemi:('.' hemisphere)? _ '=' _ expression:expression {
    identifier = identifier.toLowerCase();
    addsymbol(identifier,expression,hemi);
    if (hemi == null)
        symbol_todisplay[identifier] = true;
    else
        symbol_todisplay[identifier] = hemi[1];
}

blocked_assignment = identifier:identifier hemi:('.' hemisphere)? _ '|=' _ expression:expression {
	addsymbol(identifier.toLowerCase(),expression,hemi);
}


/// expressions ////////////////////////////

expression = disjunction

disjunction = head:term tail:(w ('or'/'+'/'|') w  term)* {
if (tail.length == 0)
	return head;
else
	return function(side) { return tail.reduce(function(result,element) {
    return FiberQuery.union(result,element[3](side));
    },head(side)); }

}

term = head:factor tail:(w ('and'/'*'/'&') w factor)* { 
if (tail.length == 0)
	return head;
else
	return function(side) { return tail.reduce(function(result,element) {
    return FiberQuery.intersect(result,element[3](side));
    },head(side)); }
}


factor =  exclusion / negation / atom

exclusion = head:atom tail:(w NOTIN w atom)+ {
	return function(side) { return tail.reduce(function(result,element) {
    return FiberQuery.diff(result,element[3](side));
    },head(side)); }
}

NOTIN = 'not in' / '-'

negation = ('not'/'-')  factor

atom = parenthesised_expression / function_evaluation / literal
parenthesised_expression = '(' _ expression:expression _ ')'  { return expression }

function_evaluation = name:function_name expression:parenthesised_expression {
	return function(side) { return FiberQuery[name](expression(side)); }
}
literal = number:number 
        / identifier:identifier hemi:('.' hemisphere)? {
                if (symbol_table[identifier.toLowerCase()] == undefined)
                {
                          var info = location();
                          if (info)
                            error("undefined identifier: " + identifier + " (line:" + info.end.line + " col:" + info.end.column +")");
                }
                identifier = identifier.toLowerCase();
                if (hemi != null)
                {
                    if (hemi[1] == "left")
        				return function() { return symbol_table[identifier]('left') };
                    else if (hemi[1] == "right")
        				return function() { return symbol_table[identifier]('right') };
                    else if (hemi[1] == "side")
                    	return symbol_table[identifier];
                    else if (hemi[1] == "opposite")
                    	return invert(symbol_table[identifier]);
                 }
                else
                	return symbol_table[identifier];  
                }
        / text:text 


number "number" = num:([0-9]+)  { return  function() {
								  if (typeof num == "object")
        								 num = num.join("");
                                  console.log(num);
                                  return FiberQuery.traversing(num);} }

text = "'"text:[^\']+"'" / '"'text:[^"]+'"' { 
    var set = undefined;
     text = text.join("");
    for (var x in symbol_table) 
    {
       if (x.search(RegExp(text.toLowerCase())) != -1)
          set =  FiberQuery.union(symbol_table[x](),set);
    }
    if (set == undefined)
	  {
		  var info = location();
		  if (info)
			  error("no match for " + text + " (line:" + info.end.line + " col:" + info.end.column +")");

	  }
    return function() { return set; };
}

string = [a-zA-Z0-9_\-\.\*\+]+

identifier = name:name {
return name;
}

name = [a-zA-Z_][a-zA-Z0-9_]* { return text() }

function_name =  "anterior_of"
              /  "posterior_of"
              /  "superior_of"
              /  "inferior_of"
              /  "lateral_of"
              /  "medial_of"
              /  "endpoints_in"
              /  "both_endpoints_in" 
              /  "only" 


