<?php
//ini_set(error_reporting(0));



// ======================================================================================
// Script can be called as ajax or from command line.
// Command line inputs are
//    -pacsid <targetpacs>, id as in conf
//    -project <project>
//    -fileID <can be SQL fileID or UNRESOLVED absolute filepath
//    -seriesNumber 
//
//    -seriesDescription <optional, filename is used as default>
//
// Multiple files can be sent with fileIDs as a list. seriesNumbers will be set increasing from the given value
// ======================================================================================



if(isset($argv) && sizeof($argv) > 1)
{
  $GLOBALS['fromcommandline'] = true;
  $_GET = parseInputKeyValue($argv);
  $_GET['command'] = "sendtopacsByFileID"; // default command

}
else
{
  if(sizeof($_POST)>0) {$_GET = array_merge($_GET,$_POST);}
  $GLOBALS['fromcommandline'] = false;
}


$GLOBALS['dicomsend_log'] = "";
$GLOBALS['dicomsend_storescu'] = "";
$GLOBALS['progress']   = 0;






require_once(dirname(__FILE__) . '/settings.php');  // __FILE__ necessary for command line calls
require_once(dirname(__FILE__) . '/routes.php');



// ======================================================================================
// ============= Main function
// ======================================================================================
$err = false;
$ret = false;
try
{
    addLogEntry("========================================== DICOMSEND LOG ==============================================", "", 0);
    if(isset($_GET['command']))
    {
      if($_GET['command'] == "sendtopacsByFileID")
      {
         $ret = sendtopacsByFileID();
      }
      else
      {
         addLogEntry("No command selected", "",1);
      }
    }
}
catch (Exception $e) 
{
   $err = "<b>Error during DICOMSEND</b><br> <span style='font-size:15px;text-align:left;'> Last Message: ". $e->getMessage() . "<span>\n";
   addLogEntry($err, "",0);
}

// ======================================================================================
// ============= generate the output of this function
// ======================================================================================

if($GLOBALS['fromcommandline'])
{
   echo "==============  command line call summary ============== \n";
   echo $GLOBALS['dicomsend_log'];
   echo $GLOBALS['dicomsend_storescu'];
}
else // called from webinterface
{
   $out = array();
   $out['dicomsend_log'] =  mb_convert_encoding($GLOBALS['dicomsend_log'], "HTML-ENTITIES", "UTF-8");     

   $out['dicomsend_err'] =  $err!=false; // & $ret=false); // ret is system command return, 0 means true in this case
   $out['dicomsend_err_lastmessage'] = $out['dicomsend_err']?$err:"";

   ob_clean();
   echo "NORA_PROGRESS_SEPARATOR";
   echo json_encode($out);  
}
 



// ======================================================================================
// ============= addLogEntry
// ======================================================================================
function addLogEntry($entry, $res="", $ret=0, $who="", $progress=0)
{
  $GLOBALS['dicomsend_log'] .= $entry;
  // flush the current step to ajax immediately
  if(!$GLOBALS['fromcommandline'] & $entry !== "" & substr($entry,0,2) != "==") 
  {
     echo "NORA_PROGRESS_SEPARATOR";
     $GLOBALS['progress'] += $progress;
     echo(json_encode(array('status' => $entry, 'progress' => $GLOBALS['progress'] / 5 )));
     echo "\n";
     flush(); ob_flush(); ob_clean();
     usleep(50);
     flush(); ob_flush(); ob_clean();
  }


    if(is_array($res)) 
      if(sizeof($res) > 0)
        $res =  implode("\n\t", $res);
      else
        $res ="";
  if($ret==0)
  {
    $GLOBALS['dicomsend_log'] .= "OK";
  }
  if($res !== "")
  {
    $GLOBALS['dicomsend_log'] .= "\n\t" . $res;
  }
  $GLOBALS['dicomsend_log'] .= "\n";

  if($ret !== 0 & $ret != -1)
  {
       throw new Exception("Error in `$who`.");
  }

}





// ======================================================================================
// ============= Called from WEBINTERFACE: 
// ======================================================================================

function sendtopacsByFileID()
{
    // this function is normally called from the webinterface, but can also be called from command line via
    if( $GLOBALS['fromcommandline'] )
    {
      $options = array();
      foreach($_GET as $key => $val)
      {
        $options[$key]  = $val;
      }
    }
    else
    {
      $options = json_decode( $_GET['json'], true);
      if(!$options)
         throw new Exception(" Error during parsing input json:" . print_r($_GET['json']) ) ;
    }

    if(!isset($_GET['project']))
         throw new Exception("Input missing: project");
    if(!isset($options['fileID']))
         throw new Exception("Input missing: fileID");
    if(!isset($options['seriesNumber']))
         throw new Exception("Input missing: seriesNumber");

    $fileIDList = explode(',' ,$options['fileID']);
    $seriesNumList = explode(',' ,$options['seriesNumber']);
    if(sizeof($fileIDList) > 20)
         throw new Exception("You selected more than 10 files. <br>For security reasons, a maximum of 20 files is permitted per job. <br>Please select less files and try again.");
    
    foreach($fileIDList as $k => $fid)
    {
      $options_temp = $options;
      $options_temp['fileID'] = $fid;
      $options_temp['seriesNumber'] = $seriesNumList[$k];
      sendtopacsByFileID_single($options_temp);
    }  


}


function sendtopacsByFileID_single($options)
{
    require_once(dirname(__FILE__) . '/mydbconnection.php');


    $query = "SELECT * FROM files WHERE id = $options[fileID]";
    $db = mydbconnection( $_GET['project'] );
    if( ($a = $db->query($query)) == false)      {$db->close(); throw new Exception(" Database error:  $db->error");}
    if( ($row = $a->fetch_assoc()) == false)       {$db->close(); throw new Exception(" File not found:  $options[fileID]") ;}
    $fileToSend = $row;

    // try to get a dicom or mat file from database. Only works if registered there. Should be the standard in future
    $query = "SELECT * FROM files WHERE patients_id = '$fileToSend[patients_id]' AND  studies_id= '$fileToSend[studies_id]' AND subfolder like 'DICOM' ";
    if( ($a = $db->query($query)) == false)      {$db->close(); throw new Exception(" Database error:  $db->error");}
    if( ($row = $a->fetch_assoc()) == false)       
    {
		//{$db->close(); throw new Exception(" No dicom header file found") ;}
		$fileDCMHeader = $row;
	}
	else
	{
	    $fileDCMHeader = false;
	}
		
    
    
    // try to get the header from DICOM folder space
    if($fileDCMHeader == false)
    {
        $query = "SELECT * FROM studies WHERE patients_id = '$fileToSend[patients_id]' AND  studyid= '$fileToSend[studies_id]' ";
        if( ($a = $db->query($query)) == false)      {$db->close(); throw new Exception(" Database error:  $db->error");}
        if( ($row = $a->fetch_assoc()) == false)       {$db->close(); throw new Exception(" No dicom header file found") ;}
        
       // print_r($row);
        $dicomheaderpath = resolve_Route($row['StudyPath'] . "/DICOM/");
        if(!is_dir($dicomheaderpath)){  $db->close();  throw new Exception("The DICOM path $dicomheaderpath does not exist or is not readable."); }

        $files  = scandir($dicomheaderpath);
        $files = array_diff(scandir($dicomheaderpath), array('..', '.'));
        
        $hdrFile = false;
                
        $a = preg_grep("/dcm$/", $files);  // look for dcm files first and take last match
        if(sizeof($a)>0)
        {     
            $hdrFile = $dicomheaderpath  . "" . end($a); 
        }

        if(!$hdrFile)  // no dicom file found, try mat
        {
            $a = preg_grep("/mat$/", $files); // look for mat files and take last match
            if(sizeof($a)>0)
            { 
              $hdrFile = $dicomheaderpath  . "" . end($a);
            }
        }

        if(!$hdrFile)
			{ $db->close(); throw new Exception("No dicom header file found inside `$dicomheaderpath`.");}


        $fileDCMHeader = array('FilePath'=> $hdrFile);
    }
    
    
    // only a mat file found, must convert the mat file to a dicom ... ufffs
    if( strpos($fileDCMHeader['FilePath'], '.mat') > 0)
    {
        $matlabcmd  =" disp('start converting'); filename = '$fileDCMHeader[FilePath]';  hdr = load(filename);     filename = [filename(1:end-3), 'dcm'];   disp(['converting  to ' filename]);  dicomdict('set', fullfile('".DPXROOT."/src/matlab/matlab_dicom_dict.txt')); dicomwrite(0 , filename, hdr ,'WritePrivate',1);   exit;   ";
        $cmd = "matlab -nodesktop -nosplash -r \"$matlabcmd\"";
        exec($cmd . " 2>&1 ", $res, $ret);  
        addLogEntry("Converting mat to dicom .... ", $res, $ret);
        $fileDCMHeader['FilePath'] = substr($fileDCMHeader['FilePath'], 0, sizeof($fileDCMHeader['FilePath'])-4) . 'dcm';
    }
   
         
    $ret = sendBMPToPacs(resolve_Route($fileToSend['FilePath']), $fileDCMHeader['FilePath'], $options);
    
    /***********************************************************
     tag with "PACS" when successfully sent
	***********************************************************/
    if($ret == 1)
    {
		$tag = "PACS";
		if($fileToSend['Tag'] == "")
		{
			$fileToSend['Tag'] = '/' . $tag . '/';
		}
		else
		{
			if (strpos($fileToSend['Tag'], '/' . $tag . '/') === false)
			{
				$fileToSend['Tag'] = $fileToSend['Tag'] . $tag . '/';
			}
		}
		$db->query("UPDATE files SET Tag='".$fileToSend['Tag']."' WHERE ID=" . $options['fileID'] );
    }
	
	$db->close();
    return $ret;

}


// ======================================================================================
// ============= test
// ======================================================================================

function runTest()
{
    // php /DPX/htdocs/dicomfuns.php image=/home/elias/data/DICOMCREATE/Kellner/KStroke_report.png  dicomheaderpath=/home/elias/data/DICOMCREATE/Kellner/DICOM/ 
    $dicomheaderpath = '/home/elias/data/DICOMCREATE/Kellner/DICOM/';
    $image 	     = '/home/elias/data/DICOMCREATE/Kellner/KStroke_report.png';

    $options = array( 
		      "seriesNumber" => "806", 
		      "seriesDescription" => "KStroke_report", 
		      "addWatermark" => true,
		      "pacsid" => 'uklfr'
		     );

    $ans = sendBMPToPacs($image, $dicomheaderpath, $options);
    return $ans;
}






// ======================================================================================
// ============= sendBMPToPacs
// ======================================================================================
function sendBMPToPacs($image, $dicomheaderpath, $options)
{


  if( !is_file($image) ) // so far, only allow sending of single files. for also folders use file_exists
  {
      throw new Exception("The file `$image` was not found.");
  }
  
  ///////////////// get a single header file from the dicom header path
  if(!file_exists($dicomheaderpath))
          throw new Exception("Dicom folder not found: `$dicomheaderpath`.");
  if( !is_file($dicomheaderpath) ) 
  {
      $files  = scandir($dicomheaderpath);
      $files = array_diff(scandir($dicomheaderpath), array('..', '.'));
      if(sizeof($files) == 0)
          throw new Exception("No dicom header file found inside `$dicomheaderpath`.");
      else
          $hdrFile = $dicomheaderpath  . "/" . end($files);
  }
  else
  {
      $hdrFile  = $dicomheaderpath;
  }

  if(substr($options['pacsid'],0,1) =='/' )
  	$targetIsLocalDir = 1;
  else
  	$targetIsLocalDir = 0;
  
  if(!$targetIsLocalDir)
  	securityCheck($hdrFile, $options['pacsid']);

  
  // get info about the file, omit .gz
  $temppath = str_replace('.gz', '', $image);
  $p = pathinfo($temppath);
  
  //////////// set standard options if none is set
  if(!isset($options['seriesDescription']))
     $options['seriesDescription'] = $p['filename'] ;
  if(!isset($options['seriesNumber']))
      $options['seriesNumber'] = 501;
  if(!isset($options['addWatermark']))
      $options['addWatermark'] = false;



  ////////////
  if(!isset($GLOBALS['settings']->pacs->SELF))
      throw new Exception("The pacs SELF definition is missing");
  if(!isset($GLOBALS['settings']->pacs->SELF->researchprefix))
      throw new Exception("The pacs SELF researchprefix is missing");
  
  
  if(isset($options['researchprefix']))
	$options['seriesDescription'] =   $options['researchprefix'] . $options['seriesDescription'];
  else
  	$options['seriesDescription'] =   $GLOBALS['settings']->pacs->SELF->researchprefix . $options['seriesDescription'];

  /////// where should the temp dicom files be stored ??
  if($targetIsLocalDir) // save as local folder
  {
  	 $hdr = dicominfo($hdrFile);
  	 $pp = pathinfo($image);
  	 $pn = explode('^', $hdr['PatientName'] );
  	 
     //$outsubfolder = $hdr['PatientID'] . "_" . $hdr['StudyDate'] . '_' . $pp['filename'];  	
  	 //$outsubfolder = $pn[0]  . '/' . $options['researchprefix'] ;  	 
     //$outsubfolder = $pn[0]  . '/' . $options['researchprefix'] . $pp['filename'];

     $outsubfolder = $options['seriesNumber'];

  	 $outpath = $options['pacsid'] ."/" . $outsubfolder ."/";
  }
  else
  {
	  $pathid = str_replace(' ','_',str_replace(".", '' , str_replace("/", '', $image) ));
	  $pathid .= str_replace(microtime(true), '.', '');
	  $outpath =  dirname(__FILE__) . "/../var/export/tmp_exported_dicoms/$pathid";
  }
  


  if(mymkdir($outpath) == false)
      throw new Exception("Could not create the output directory $outpath.");

  
  
  
  //// bitmaps and niftis are converted in different ways ....
      
  ///////////////////////// Bitmap
  if( in_array($p['extension'], array('png', 'jpeg', 'bmp', 'jpg')  ) ) 
  {
      $dcmfiles = convert_bitmap_to_dicom($image, $outpath, $hdrFile, $options);
  }
  ///////////////////// Nifti
  elseif( in_array($p['extension'], array('nii')  )  || strpos($image, 'nii.gz' ) !== false ) 
  {
      $dcmfiles = convert_nifti_to_dicom($image, $outpath, $hdrFile, $options);
  }
  else
  {
      throw new Exception("The file $image has an invalid/unsupported extension");
  }
  
  ////////////////////  send to pacs
  if( $targetIsLocalDir )
  {
  	 addLogEntry("============================================================");
  	 addLogEntry("Not sending dicoms, but writing to local folder: " . $outpath . "\n");
  	 addLogEntry("============================================================");
  	 $ret = true;
  }
  else
  {
  	$ret = sendDICOMToPACS($dcmfiles, $options['pacsid']);	
  }
  
  return $ret;
  
}

// ======================================================================================
// ============= Bitmap specific function
// ======================================================================================

function convert_bitmap_to_dicom($in, $outpath, $hdrFile, $options)
{

  $p = pathinfo($in);

        // so far, only the bmp extension will work!!! storescu problem ??
      //$extension = "bmp";
      // new approach: jpg works acually better, with dcmdjpeg after writing the dicom, dicom-decode-jpeg, see below
      $extension = "jpg";
      // ************ create a temporary bitmap
      //$tempbmp = $p['dirname'] ."/". $p['filename'] . "_temp.$extension";
      $tempbmp = $outpath ."/". $p['filename'] . "_temp.$extension";
      $cmd = "gm convert $in  $tempbmp  ";
      exec($cmd . " 2>&1 ", $res, $ret);  
      addLogEntry("Converting to temporary bitmap .... $cmd ", $res, $ret, "ImageMagick convert");
      // ************ add a watermark if desired
      if( $options['addWatermark'] )
      {
	  addWatermark($tempbmp );
      }

  
  $p = pathinfo($tempbmp);
  $outdicom = $outpath ."/". $p['filename'] . '.dcm' ;

  
  // **************  Modify inportant tags directly in img2dcm *************
  $modify = 
    ' -k "SeriesDescription='. $options['seriesDescription']. '" ' .
    ' -k "ProtocolName=' .$options['seriesDescription']. '" ' .
    ' -k "SeriesNumber=' .$options['seriesNumber']. '" ' .
    ' -k "AcquisitionNumber=1" ' .
    ' -k "RescaleSlope=1" ' .
    ' -k "RescaleIntercept=0" ' .
    ' -k "RescaleType=US" ' .
    ' -k "WindowCenter=128" ' .
    ' -k "WindowWidth=255" '
    ;
  
  // **************  bitmap to  dicom *************
  $BMP = $p['extension']=="bmp"?"-i BMP":"";
  //$cmd = "img2dcm $BMP --dataset-from   $hdrFile  $modify  $tempbmp $outdicom";
  $cmd = "img2dcm $BMP --study-from   $hdrFile  $modify  $tempbmp $outdicom";
  exec($cmd . " 2>&1 ", $res, $ret);  
  addLogEntry("Converting img2dcm ... ", $res, $ret, "img2dcm");
  unlink($tempbmp); // delete the temp bitmap again


  $hdr = dicominfo($hdrFile);

  ////////////// copy header info from the template file
  $str = "";
  foreach($hdr as $key => $val) 
    $str .= " -i $key=\"$val\"";

  if(!$BMP)
    $str .= " -e 0028,2110 ";
  
  $cmd = "dcmodify --no-backup  $str $outdicom";
  exec($cmd . " 2>&1 ", $res, $ret);

/*
  if(!$BMP)
  {
    //remove the tag saying the image underwent lossy compression and others
    $cmd = "dcmodify --no-backup -e 0028,2110 -e FlipAngle -e RepetitionTime -e EchoTime";
    exec($cmd . " 2>&1 ", $res, $ret);
  }
  */


  // decompress the jpeg to uncomressed dicom, if not derived from a BMP
  if(!$BMP)
  {
    $cmd = "dcmdjpeg $outdicom $outdicom";
    $res = array();
    exec($cmd . " 2>&1 ", $res, $ret);
    addLogEntry("Decompressing written dicom using dcmd2jpeg ... ", $res, $ret, "dcmdjpeg");
  }   
  
  $hdr = dicominfo($hdrFile);
  
  $dcmfiles = array($outdicom);
  set_new_dicom_header($dcmfiles, $hdr, $options);
  return $dcmfiles;
}

// ======================================================================================
// ============= NIFTI specific function
// ======================================================================================

function convert_nifti_to_dicom($in, $outpath, $hdrFile, $options)
{
  $hdr = dicominfo($hdrFile);

  ////////////// copy header info from the template file
  $str = "";
  foreach($hdr as $key => $val) 
    $str .= " -i $key=\"$val\"";
  //die($str ."\n \n");
  
  // create some halfways unique id for this file from its full path
  
  // run node to create the dicom
  $jsfile = dirname(__FILE__) . "/../src/node/dicom/dicomwrite.js";

  $nodepath = "nodejs";
  $nodepath = dirname(__FILE__) . "/../bin/node";
  $cmd = "$nodepath  '$jsfile' -source '$in' -target '$outpath'";

  addLogEntry("Converting NIFTI to DICOM ... ", "", -1, "nodjs nifti to dicom", 1);
  exec($cmd . " 2>&1 ", $res, $ret);
  //system($cmd);// . " 2>&1 ", $res, $ret);
  addLogEntry("", $res, $ret, "nodejs nifti to dicom");
  
  
  $dcmfiles = array_values(array_diff(scandir($outpath), array('..', '.')));
  if(sizeof($dcmfiles) == 0)
	throw new Exception("ERROR: No files created or found in $outpath");

  foreach($dcmfiles as &$file) // cat the base path
      $file = $outpath . "/" . $file;
  
  $cmd = "dcmodify --no-backup  $str $outpath/*.dcm";
  exec($cmd . " 2>&1 ", $res, $ret);
//echo($str); die()
//   $hdr = dicominfo("$outpath/*.dcm");var_dump($hdr); die();

  set_new_dicom_header($dcmfiles, $hdr, $options);
  
  return $dcmfiles;

}



// ======================================================================================
// ============= convert_image_to_dicom
// ======================================================================================

function set_new_dicom_header($dcmfiles, $hdr, $options)
{
  // note on series number and seriesUID:
  // the pacs will accept series with same seriesNumber (for example 8881), but different seriesUID.
  // two series will show up with same number. BUT: first series will have ALL Elements!!!
  // so: no good idea!

  ////////////   MODIFY UIDs (SeriesInstanceUID and SOPInstanceUIDs)  
  // get the old ID
  
  // OUR DICOM ROOT (length 23)
  $DICOMROOT = "1.2.276.0.76.3.1.248.3";
  
  // cat microtime and a randnum (length 16)
  $DICOMROOT .= ("." . microtime(true) );
  
  // add a random number (length 3)
  $DICOMROOT .= ("." . mt_rand(1,999) );

  /* OOOOOLD creation of UIDS
  $SUID_base = substr($hdr['SeriesInstanceUID'], 0,64 - strlen($options['seriesNumber']) ); // final suid can only have 64 chars, reserve seriesNumber 
  if( strlen($SUID_base) < 20 )
       throw new Exception("No valid SeriesInstanceUID in template header file");
  
  $SOPUID_base = substr($hdr['SOPInstanceUID'], 0,64 - 16 - 4); // final suid can only have 64 chars, reserve 16 for microtime + a file instance counter
  if(strlen($SOPUID_base) < 20)
       throw new Exception("No valid SOPInstanceUID in template header file");

  // simply cat a new series number, suid is the same for all instances!
  // suid is based on series number ... maybe make it base on a time stamp ...?
  $SUID = $SUID_base . "." . $options['seriesNumber'];
  $SUID = substr($SUID, 0,64);  // for security: max lenght 64
  */
  
  // NEW: Series is root cat 1
  $SUID = $DICOMROOT . ".1";
  
  addLogEntry("Modifying DICOM header  ... ", "", -1, "dcmodify", 1);

  for($k=0; $k<sizeof($dcmfiles); $k++)
  {
      
      /* old
      // SOPUID must increment, this is the only significant id on file level! (microtime might be enough, but we use also an additional counter)
      $SOPUID = $SOPUID_base . "." . microtime(true) . strval($k);
      $SOPUID = substr($SOPUID, 0,64); 
      */
		
	  // take dicomroot and start with 2 for SOPS
	  $SOPUID = $DICOMROOT . "." . ($k+2);
    
      $cmd = "dcmodify --no-backup  -i SeriesInstanceUID=$SUID  \
				    -i SOPInstanceUID=$SOPUID  \
				    -i SeriesDescription=$options[seriesDescription] \
				    -i ProtocolName=$options[seriesDescription] \
				    -i SeriesNumber=$options[seriesNumber]  \
				    -i ImageComments=\"RESEARCH IMAGE: Not for Diagnostic Use!\" \
				    $dcmfiles[$k]";

      exec($cmd . " 2>&1 ", $res, $ret);
   }
 
  addLogEntry("Setting new Instance UIDS and series information ... ", $res, $ret, "dcmodify");

  $res[] =   "OLD DICOM file had the following IDs:";
    if(!isset($hdr['seriesDescription'])) $hdr['seriesDescription'] = 'no_description_in_template';
    if(!isset($hdr['seriesNumber'])) $hdr['seriesNumber'] = 'no_number_in_template';
  $res[] =   "\tseriesDescription : $hdr[seriesDescription]";
  $res[] =   "\tseriesNumber      : $hdr[seriesNumber]";
  $res[] =   "\tSeriesInstanceUID : $hdr[SeriesInstanceUID]";
  $res[] =   "\tSOPInstanceUID    : $hdr[SOPInstanceUID]";
  $res[] =   "\tSOPClassUID    : $hdr[SOPClassUID]";

  $res[] =   "NEW DICOM file has the following IDs:";
  $res[] =   "\tseriesDescription : $options[seriesDescription]";
  $res[] =   "\tseriesNumber      : $options[seriesNumber]";
  $res[] =   "\tSeriesInstanceUID : $SUID";
  $res[] =   "\tSOPInstanceUID    : $SOPUID";

  addLogEntry("Setting new Instance UIDS and series information ... ", $res, $ret, "dcmodify");

  
  return $dcmfiles;
  
}




// ======================================================================================
// ============= sendDICOMToPACS
// ======================================================================================

function sendDICOMToPACS($dcmfiles,$targetpacs)
{
  addLogEntry("=======================================================================================================", "", -1);
  addLogEntry("DICOMSEND ... ", "", -1);
 
  // multiple dicoms: send the whole folder
  if(sizeof($dcmfiles) > 1)
  {
    $p = pathinfo($dcmfiles[0]); 
    $sendpath = $p['dirname'] . '/';
    $sendoptions = ' --scan-directories ';
  }
  else  
  {
    $sendoptions  = '';
    $sendpath = $dcmfiles[0];
  }


  if( !file_exists($sendpath) ) // so far, only allow sending of single files. for also folders use file_exists
        throw new Exception("The dicomfile/folder to send: $sendpath was not found.");

  

  // check again with this final dicom file to send
  securityCheck($dcmfiles[0], $targetpacs);

  
  // get the pacs list
  $pacs = $GLOBALS['settings']->pacs;
  
  // this is already checked in securityCheck
  /*
  if( !isset( $pacs->$targetpacs ))
  {
  		delete_after_sending($dcmfiles);      
        throw new Exception("ERROR: The pacsid `$targetpacs` is not in my list.");
  }
  */

  $myAET = $pacs->SELF->aetitle;
 
  $targetAET = $pacs->$targetpacs->aetitle;
  $targetIP = $pacs->$targetpacs->ip;
  $targetPORT = $pacs->$targetpacs->port;
  $seriousSend = $pacs->$targetpacs->seriousSend;
  
      
  //**********  run the send DICOM **********
  if(!$myAET | !$targetAET | !$targetIP | !$targetPORT )
  {
		delete_after_sending($dcmfiles);      
        throw new Exception("ERROR: Some of the necessary PACS information was corrupt in\n ".  $cmd);
  }

  $cmd = "storescu $sendoptions --verbose  --aetitle $myAET --call $targetAET  $targetIP  $targetPORT    $sendpath ";

  addLogEntry("---------------------------------------------------------------", "", -1);
  addLogEntry("sending dicoms ... ", "", -1, 'storescu', 1);

  if($seriousSend == false)
  {
      addLogEntry("!!! IN SIMULATION MODE, executing " ,$cmd, -1);
      
      delete_after_sending($dcmfiles);
      
      throw new Exception("This PACS is in simulation mode. Check the log to see if everything went all right.");
      return true;
  }
  else
  {
      addLogEntry("Sending dicoms in serious mode to $targetpacs using --aetitle $myAET --call $targetAET  $targetIP  $targetPORT   ..." ,"", -1);
      exec($cmd . " 2>&1" , $res, $ret);
      $out['dicomsend_protocol'] = $res;

	 
	  delete_after_sending($dcmfiles);      
	 
	 
      // check what storescu told us.
      {
         addLogEntry("STORESCU messages :",  $res, $ret, "STORESCU"); // do not abort here ? must set $ret to 0
         // save the send protocol also in another variable
         
         // this is actually not called if $ret =1 ...
         $GLOBALS['dicomsend_storescu_protocol'] = $res;
         //die(var_dump($res));
         //$arr = explode($res, "\n");
         if(  strpos( implode($res), "TCP Initialization Error: Connection refused") > 0)
         {
             throw new Exception("STORESCU connection error. Could not connect to the remote PACS.");
         }
         else // seemed to work nicely
         {
             return true;
         }
      }
      

  }
  

}


function delete_after_sending($dcmfiles)
{
	addLogEntry("---------------------------------------------------------------", "", -1);
    addLogEntry("deleting files after sending ... ", "", -1, 'storescu', 1);
	addLogEntry("---------------------------------------------------------------", "", -1);
    
    $p = pathinfo($dcmfiles[0]); 
    $sendpath = $p['dirname'] . '/';
    array_map('unlink', glob("$sendpath*.dcm"));
    rmdir($sendpath);
}



function securityCheck($hdrfile, $targetpacs)
{
	
	// for / pacs, send dicoms to a local folder
// 	if(substr($pacs = $GLOBALS['settings'],0,1) =='/' )
// 		return true;

    $pacs = $GLOBALS['settings']->pacs;

	if( !isset( $pacs->$targetpacs ))
		throw new Exception("ERROR: The pacsid `$targetpacs` is not in my list.");

    $hdr = dicominfo($hdrfile);

    if( empty($hdr)  ) 
         throw new Exception("The DICOM header of file $hdrfile was not readable.");
    if( strlen($hdr['SeriesInstanceUID']) < 20 )
	throw new Exception("No valid SeriesInstanceUID in file(s)");
    if( strlen($hdr['StudyInstanceUID']) < 20 )
	throw new Exception("No valid StudyInstanceUID in file(s)");
    if( strlen($hdr['PatientID']) < 4 )
	throw new Exception("No valid PatientID in file(s)");
    
    
    $allowed = false;
    if(isset( $pacs->$targetpacs->DCMFilterInstitutionNames))
    {
      if( $pacs->$targetpacs->DCMFilterInstitutionNames[0] == "*" | in_array( $hdr['InstitutionName'],  $pacs->$targetpacs->DCMFilterInstitutionNames  ) )
        $allowed = true;
      else
        throw new Exception("Security Error: The DICOM InstitutionName <br>`". $hdr['InstitutionName'] ."`<br> is  not in the list for this PACS  <br>`" . implode(',', $pacs->$targetpacs->DCMFilterInstitutionNames) . "`.");
    }
    if(isset( $pacs->$targetpacs->DCMFilterIssuerOfPatientID))
    {
      if($pacs->$targetpacs->DCMFilterIssuerOfPatientID[0] == "*" | in_array( $hdr['IssuerOfPatientID'],  $pacs->$targetpacs->DCMFilterIssuerOfPatientID  ))
        $allowed = true;
      else
        throw new Exception("Security Error: The DICOM IssuerOfPatientID `". $hdr['IssuerOfPatientID'] ."` is  not in the list for this PACS  `" . implode(',', $pacs->$targetpacs->DCMFilterIssuerOfPatientID) . "`.");
    }
    if(!$allowed)
    	  throw new Exception("Security Error: No InstitutionName or IssuerOfPatientID List set for this PACS. Cannot proceed.");
}


// ======================================================================================
// ============= dicominfo
// ======================================================================================

function dicominfo($file, $echo_result=0)
{
  $tags = array();
  $tags["0008,0005"] = 'SpecificCharacterSet'; 

// we cannot modify 0002,0002 tags, this will lead to error!
// so leave out here!!
//  $tags["0002,0002"] = 'MediaStorageSOPClassUID'; 

  $tags["0008,003e"] = 'SeriesDescription'; 
  $tags["0008,0018"] = 'SOPInstanceUID'; 
  $tags["0008,0080"] = 'InstitutionName'; 
  $tags["0008,0081"] = 'InstitutionAddress'; 

  $tags["0008,0020"] = 'StudyDate'; 
  $tags["0008,0021"] = 'SeriesDate'; 
  $tags["0008,0030"] = 'StudyTime'; 
  $tags["0008,0031"] = 'SeriesTime'; 
  $tags["0008,0032"] = 'AcquisitionTime'; 

  
  // has invalid values ...?
  $tags["0008,0050"] = 'AccessionNumber'; 
  $tags["0008,0016"] = 'SOPClassUID'; 

  $tags["0008,0060"] = 'Modality';      

  $tags["0008,1030"] = 'StudyDescription'; 

  $tags["0018,1000"] = 'DeviceSerialNumber'; 
  $tags["0018,0015"] = 'BodyPartExamined'; 
  $tags["0018,5100"] = 'PatientPosition'; 

  $tags["0020,0011"] = 'SeriesNumber'; 
  $tags["0020,000e"] = 'SeriesInstanceUID'; 
  $tags["0020,000d"] = 'StudyInstanceUID'; 
  $tags["0020,0010"] = 'StudyID'; 


  $tags["0010,0010"] = 'PatientName'; 
  $tags["0010,0020"] = 'PatientID'; 
  $tags["0010,0030"] = 'PatientBirthDate'; 
  $tags["0010,0040"] = 'PatientSex'; 
  $tags["0010,1005"] = 'PatientBirthName'; 
  $tags["0010,1010"] = 'PatientAge'; 
  $tags["0010,1030"] = 'PatientWeight'; 
  
  $tags["0010,0021"] = 'IssuerOfPatientID'; 
      
  
  

  $t = implode(' +P ', array_keys($tags));
  //$cmd = "dcmdump  +P $t " . $file;
  $cmd = "dcmdump --no-uid-names +P $t " . $file;
  
  $res = array();
  exec($cmd . " 2>&1 ", $res, $ret);
  if($echo_result)
    addLogEntry("== Reading DICOM Header of $file ... ", $res, $ret);

  $hdr = array();
  $matches = array();
  foreach($res as $row)
  {
      preg_match_all (  "/\(   (?<tag>\w{4},\w{4})   \)\s\w\w\s\[   (?<value>[^]]+)   \]\s#.*/x" , $row, $matches, PREG_SET_ORDER ); // x means igonore empty spaces
      //print_r($matches[0]);
      if(!empty($matches))
      {
	  	$hdr[$tags[$matches[0]['tag']]] = $matches[0]['value'];
      }
      else
      {
		//addLogEntry("",        "************* Warning: Tag in  " . $row . " value could not be parsed here on command line",  -1);
      }
   }
   
  //$hdr['SOPClassUID'] = '=CTImageStorage';
   
   return ($hdr);
 
}




// ======================================================================================
// ============= addWatermark
// ======================================================================================

function addWatermark($image, $text=0)
{

    # get the image size and rescale background to it
    $scale=0.8;
    $bg_sizeX = 100;
    $bg_sizeX = exec("identify -format '%w' $image");
    $bg_sizeY = exec("identify -format '%h' $image ");

    $bg_size = $bg_sizeX*$scale ."x" .  $bg_sizeY*$scale ;

    if($text==0)
    {
      $text = "R E S E A R C H   O N L Y ";
      $text = "N O R A    I M A G I N G";
      $text = "N O R A    I M A G I N G     -    R E S E A R C H   O N L Y";
    }
    $rotation=0;
    $opacity=0.6;
    $tile="";
		  
		    
    $cmd = "convert -background 'rgba(50,50,50,0.5)' -font helvetica -pointsize 29 -fill white label:'$text' miff:- | composite $tile -gravity south -geometry $bg_size+0+5  -  $image  ". $image ." ";
    exec($cmd . " 2>&1 ", $res, $ret);  
    addLogEntry("Adding a watermark ....", $res, $ret);

}







// mkdir can throw several errors. Permission errors are important and must be filterd!
// Only abort if no permissions ( error can also occur if file exists)
function mymkdir($targetpath)
{

  if(!@mkdir( $targetpath , 0775,true ) )
  {
    $err = error_get_last();
    if(strpos($err['message'], 'ermission') !== false )
    {
      return false;
    }
  }
  exec("chmod g+wsx $targetpath 2>&1");

  return true;

}





function parseInputKeyValue($argv)
{
    $json = array();
    for($k=1; $k<sizeof($argv)-1; $k+=2)
    {
      $json[substr( $argv[$k], 1, strlen($argv[$k]) )] =   $argv[$k+1];
    }
     //$inputsummary = "=========== END OF INPUT   ===================\n";
     //$inputsummary .= print_r($json);
     //echo $inputsummary;
     return $json;

}


?>

