"use strict";
if (typeof module != "undefined")
{
	var KissFFTModule = require('./kissFFT.js');
}

var kissFFTModule = KissFFTModule({});

var kiss_fftr_alloc = kissFFTModule.cwrap(
    'kiss_fftr_alloc', 'number', ['number', 'number', 'number', 'number' ]
);

var kiss_fftr = kissFFTModule.cwrap(
    'kiss_fftr', 'void', ['number', 'number', 'number' ]
);

var kiss_fftri = kissFFTModule.cwrap(
    'kiss_fftri', 'void', ['number', 'number', 'number' ]
);

var kiss_fftr_free = kissFFTModule.cwrap(
    'kiss_fftr_free', 'void', ['number']
);

var kiss_fft_alloc = kissFFTModule.cwrap(
    'kiss_fft_alloc', 'number', ['number', 'number', 'number', 'number' ]
);

var kiss_fft = kissFFTModule.cwrap(
    'kiss_fft', 'void', ['number', 'number', 'number' ]
);

var kiss_fft_free = kissFFTModule.cwrap(
    'kiss_fft_free', 'void', ['number']
);

var FFT = function (size) {

    this.size = size;
    this.fcfg = kiss_fft_alloc(size, false);
    this.icfg = kiss_fft_alloc(size, true);
    
    this.inptr = kissFFTModule._malloc(size*8 + size*8);
    this.outptr = this.inptr + size*8;
    
    this.cin = new Float32Array(kissFFTModule.HEAPU8.buffer, this.inptr, size*2);
    this.cout = new Float32Array(kissFFTModule.HEAPU8.buffer, this.outptr, size*2);
    
    this.forward = function(cin) {
	this.cin.set(cin);
	kiss_fft(this.fcfg, this.inptr, this.outptr);
	return new Float32Array(kissFFTModule.HEAPU8.buffer,
				this.outptr, this.size * 2);
    };
    
    this.inverse = function(cpx) {
	this.cin.set(cpx);
	kiss_fft(this.icfg, this.inptr, this.outptr);
	return new Float32Array(kissFFTModule.HEAPU8.buffer,
				this.outptr, this.size * 2);
    };
    
    this.dispose = function() {
	kissFFTModule._free(this.inptr);
	kiss_fft_free(this.fcfg);
	kiss_fft_free(this.icfg);
    }
};

var FFTR = function (size) {

    this.size = size;
    this.fcfg = kiss_fftr_alloc(size, false);
    this.icfg = kiss_fftr_alloc(size, true);
    
    this.rptr = kissFFTModule._malloc(size*4 + (size+2)*4);
    this.cptr = this.rptr + size*4;
    
    this.ri = new Float32Array(kissFFTModule.HEAPU8.buffer, this.rptr, size);
    this.ci = new Float32Array(kissFFTModule.HEAPU8.buffer, this.cptr, size+2);
    
    this.forward = function(real) {
	this.ri.set(real);
	kiss_fftr(this.fcfg, this.rptr, this.cptr);
	return new Float32Array(kissFFTModule.HEAPU8.buffer,
				this.cptr, this.size + 2);
    };
    
    this.inverse = function(cpx) {
	this.ci.set(cpx);
	kiss_fftri(this.icfg, this.cptr, this.rptr);
	return new Float32Array(kissFFTModule.HEAPU8.buffer,
				this.rptr, this.size);
    };
    
    this.dispose = function() {
	kissFFTModule._free(this.rptr);
	kiss_fftr_free(this.fcfg);
	kiss_fftr_free(this.icfg);
    }
};






/*
    insert a source vector into a destination vector
    src: the vector to be inserted
    dest: the vector that src is inserted into
    n: number of elements of src to insert
    offset: the position in dest that src is inserted into
*/
function copy_into(src,dest,n,offset)
{
    for(var i = 0; i < n; i++)
    {
        dest[offset+i] = src[i];
    }
}





function fft_2d(kspace, dim, img_space,is_cplx)
{
	return  _fft_2d(kspace, dim, img_space,is_cplx,'forward');
}
function ifft_2d(kspace, dim, img_space,is_cplx)
{
	return  _fft_2d(kspace, dim, img_space,is_cplx,'inverse');
}



function _fft_2d(img, dim, img_ft, is_cplx,dir)
{




	function transpose_cplx(img,dim,img_t)
	{
		for(var i = 0; i < dim[0]; i++)
		{
			for(var j = 0; j < dim[1]; j++)
			{
				img_t[2*(j+i*dim[1])] = img[2*(i+j*dim[0])];
				img_t[2*(j+i*dim[1]) + 1] = img[2*(i+j*dim[0]) + 1];
			}
		}
	}

    //create fft objects
    var kissFFT0 = new KissFFT.FFT(dim[0]);
    var kissFFT1 = new KissFFT.FFT(dim[1]);


    //create vectors that will store intermediate steps
    var frow = new Float64Array(2*dim[0]);
    var intermed = new Float64Array(2*dim[0]*dim[1]);
    var intermed_t = new Float64Array(2*dim[0]*dim[1]);
    var fft_trans = new Float64Array(2*dim[0]*dim[1]);


    // make image complex if not already, makes it easier, only complex fft
    if(!is_cplx)
    {
        var img2 = new Float64Array(dim[0]*dim[1]*2);
        
        // not working in all nodejs versions ... so zero fill manually
        //img2.fill(0);
        for(var k=0; k< img2.length; k++) 
            img2[k] = 0;

        for(var k=0; k< img.length; k++)
            img2[2*k] = img[k];
        img = img2;
        var fft_type = 'complex';
    }

    //perform 1D fft on each of the rows of img
    var row_len = 2*dim[0];
    for(var i = 0; i < dim[1]; i++)
    {
        //fft0.simple(frow,img.slice(i*row_len,(i+1)*row_len),fft_type);
        frow = kissFFT0[dir](img.slice(i*row_len,(i+1)*row_len));
        copy_into(frow,intermed,2*dim[0],i*2*dim[0]);
    }

    //transpose the row transformed array
    transpose_cplx(intermed,dim,intermed_t);

    //fft each of the rows (which are columns from the original image)
    var col_len = 2*dim[1];
    var fcol = new Float64Array(col_len);
    for(var i = 0; i < dim[0]; i++)
    {
        //fft1.simple(fcol,intermed_t.slice(i*col_len,(i+1)*col_len),'complex');
        fcol = kissFFT1[dir](intermed_t.slice(i*col_len,(i+1)*col_len));
        copy_into(fcol,fft_trans,col_len,i*col_len);
    }

    //transpose back to original dimentions
    var dim_t = new Float64Array([dim[1], dim[0]]);
    transpose_cplx(fft_trans,dim_t,img_ft);
}






function fft_3d(kspace, dim, img_space,is_cplx)
{
	return  _fft_3d(kspace, dim, img_space,is_cplx,'forward');
}
function ifft_3d(kspace, dim, img_space,is_cplx)
{
	return  _fft_3d(kspace, dim, img_space,is_cplx,'inverse');
}


function _fft_3d(img, dim, img_ft, is_cplx,dir)
{
    //create fft objects
    var kissFFT0 = new KissFFT.FFT(dim[0]);
    var kissFFT1 = new KissFFT.FFT(dim[1]);
    var kissFFT2 = new KissFFT.FFT(dim[2]);

    //create vectors that will store intermediate steps
    var frow;
    var tmp = new Float64Array(2*dim[0]*dim[1]*dim[2]);
    
	var fac = 1/Math.sqrt(dim[0]*dim[1]*dim[2]);

    // make image complex if not already, makes it easier, only complex fft
    if(!is_cplx)
    {
        var img2 = new Float64Array(dim[0]*dim[1]*dim[2]*2);
        
        // not working in all nodejs versions ... so zero fill manually
        //img2.fill(0);
        for(var k=0; k< img2.length; k++) 
            img2[k] = 0;

        for(var k=0; k< img.length; k++)
            img2[2*k] = img[k];
        img = img2;
        var fft_type = 'complex';
    }

    //perform 1D fft on each of the rows of img
    var len = dim[0];
    for(var z = 0; z < dim[2]; z++)
		for(var y = 0; y < dim[1]; y++)
		{
			var i = y + z*dim[1]; // x + dim[0]*(y+ z*dim[1])
			frow = kissFFT0[dir](img.slice(i*2*len,(i+1)*2*len));
			for (var x = 0; x <len;x++)
			{
			   var i = y + x*dim[1] + z*dim[1]*dim[0]; 
			   img_ft[2*(i)] = frow[2*x] ;
			   img_ft[2*(i)+1] = frow[2*x+1];
			}
		}

    var len = dim[1];
    for(var z = 0; z < dim[2]; z++)
		for(var x = 0; x < dim[0]; x++)
		{
			var i = x + z*dim[0]; // y + dim[1]*(x + z*dim[0])
			frow = kissFFT1[dir](img_ft.slice(i*2*len,(i+1)*2*len));
			for (var y = 0; y <len;y++)
			{
			   var i = z + y*dim[2] + x*dim[1]*dim[2];
			   tmp[2*(i)] = frow[2*y];
			   tmp[2*(i)+1] = frow[2*y+1];
			}
		}

   var len = dim[2];
   for(var x = 0; x < dim[0]; x++)
        for(var y = 0; y < dim[1]; y++)
		{
			var i =  y + x*dim[1]; // z + dim[2]*(y + x*dim[1]);
			frow = kissFFT2[dir](tmp.slice(i*2*len,(i+1)*2*len));
			for (var z = 0; z <len;z++)
			{
			   var i = x + y*dim[0] + z*dim[1]*dim[0];
			   img_ft[2*(i)] = frow[2*z] *fac;
			   img_ft[2*(i)+1] = frow[2*z+1] *fac;
			}
		}

}





if(typeof module != "undefined")
{
	module.exports = 
	{
		FFT: FFT,
		FFTR: FFTR
	};
}
else
{
	var KissFFT = 
	{
		FFT: FFT,
		FFTR: FFTR
	};
}




function testFFT()
{
  var nii = KViewer.viewports[0].getCurrentViewer().nii;
  var sz = nii.sizes;

  
  
  for (var j = 0; j < sz[2];j++)
  {
      var tmp = new Float64Array(2*sz[0]*sz[1]);

  	  fft_2d(
  	    nii.data.slice(j*sz[0]*sz[1], (j+1)*sz[0]*sz[1]),
  	    sz,tmp,false)

      for (var k = 0; k < sz[0]*sz[1];k++)
         { tmp[2*k] = tmp[2*k]*tmp[2*k] + tmp[2*k+1]*tmp[2*k+1];
           tmp[2*k+1] = 0;}

  	  ifft_2d(tmp,sz,tmp,true);
	
      for (var k = 0; k < sz[0]*sz[1];k++)
         { nii.data[j*sz[0]*sz[1] +k ] = tmp[2*k]; };
          

  }

}

function prepare(nii)
{
	
	var voxSize = [3,3,3];
	var pad = 10;
    var new_sz = [ Math.floor(nii.sizes[0]*nii.voxSize[0]/voxSize[0]),
     				  Math.floor(nii.sizes[1]*nii.voxSize[1]/voxSize[1]),
     				  Math.floor(nii.sizes[2]*nii.voxSize[2]/voxSize[2])];
    var new_sizes = [ new_sz[0]+pad*2,
     				  new_sz[1]+pad*2,
     				  new_sz[2]+pad*2 ];
    var tot = new_sizes[0]*new_sizes[1]*new_sizes[2];

	var sx = nii.voxSize[0]/voxSize[0];
	var sy = nii.voxSize[1]/voxSize[1];
	var sz = nii.voxSize[2]/voxSize[2];

	var A = [
	[1/sx , 0 , 0 , -pad*voxSize[0] ],
	[0, 1/sy , 0  , -pad*voxSize[1] ],
	[0, 0 ,1/sz  , -pad*voxSize[2] ],
	[0 ,0 ,0, 1]];

	var new_edges = math.multiply(nii.edges,A);



	
	var fileobj1 = { content : { edges: new_edges,  bbox: undefined, voxSize: voxSize, sizes: new_sizes, pixdim: [1,voxSize[0],voxSize[1],voxSize[2]] } };
	var fileobj2 = { content : { edges: new_edges,  bbox: undefined, voxSize: voxSize, sizes: new_sizes, pixdim: [1,voxSize[0],voxSize[1],voxSize[2]] } };
	var fobj1 =  cloneNifti(fileobj1,"test1","float",1);
	var fobj2 =  cloneNifti(fileobj2,"test2","float",1);
	
 	
    for (var z = 0; z < new_sz[2];z++)
    for (var y = 0; y < new_sz[1];y++)
    for (var x = 0; x < new_sz[0];x++)
    {
		var idx = x+pad + new_sizes[0]*(pad+y) + new_sizes[1]*new_sizes[0]*(z+pad);
		var x_ = math.floor(x/sx);
		var y_ = math.floor(y/sy);
		var z_ = math.floor(z/sz);
		var idx_ = x_ + nii.sizes[0]*y_ +  nii.sizes[1]*nii.sizes[0]*z_;
		fobj1.content.data[idx] = nii.data[idx_];
		var idx_ = x_ + nii.sizes[0]*y_ +  nii.sizes[1]*nii.sizes[0]*(nii.sizes[2]-z_-1);
		fobj2.content.data[idx] = nii.data[idx_];

    }



autocorr(fobj1.content);
autocorr(fobj2.content);

    fobj1.fileID = "test1"   
    KViewer.dataManager.setFile(fobj1.fileID,fobj1);
    fobj2.fileID = "test2"   
    KViewer.dataManager.setFile(fobj2.fileID,fobj2);
    KViewer.cacheManager.update();




}


function autocorr(nii)
{
	  var sz = nii.sizes;


	  var tmp = new Float64Array(2*sz[0]*sz[1]*sz[2]);
	  var tmp2 = new Float64Array(2*sz[0]*sz[1]*sz[2]);

	  fft_3d(nii.data,sz,tmp,false)

	  var x2 = Math.floor(sz[0]/2);
	  var y2 = Math.floor(sz[1]/2);
	  var z2 = Math.floor(sz[2]/2);


	  for (var z = 0; z < sz[2];z++)
	  for (var y = 0; y < sz[1];y++)
	  for (var x = 0; x < sz[0];x++)
		 { 
		   var k = x+y*sz[0]+z*sz[0]*sz[1];
		   var xi = Math.cos(x/sz[0]*2*Math.PI);
		   var yi = Math.cos(y/sz[1]*2*Math.PI);
		   var zi = Math.cos(z/sz[2]*2*Math.PI);
		   var f = Math.pow(xi*yi*zi,4);
		   tmp[2*k] = (tmp[2*k]*tmp[2*k] + tmp[2*k+1]*tmp[2*k+1]);
		   tmp[2*k+1] = 0;
		 }

	  ifft_3d(tmp,sz,tmp2,true)


	  for (var z = 0; z < sz[2];z++)
	  for (var y = 0; y < sz[1];y++)
	  for (var x = 0; x < sz[0];x++)
		 { 
			var i = x+y*sz[0]+z*sz[0]*sz[1];
			var j = (x+x2)%sz[0]+((y+y2)%sz[1])*sz[0]+((z+z2)%sz[2])*sz[0]*sz[1];
			nii.data[i] = tmp2[2*j]/100000;

		 }

}


function rotslice(img,dest,sz,phi)
{
	var c = Math.cos(phi);
	var s = Math.sin(phi);
	for (var y = 0; y < sz[1];y++)
 		for (var x = 0; x < sz[0];x++)
 		{
			var a = Math.floor(((x-sz[0]/2)*c + (y-sz[1]/2)*s) + sz[0]/2);
			var b = Math.floor((-(x-sz[0]/2)*s + (y-sz[1]/2)*c) + sz[1]/2);
			if (a>=0 && a < sz[0] && b>=0 && b < sz[1])
			{
				dest[x+y*sz[0]] = img[a+b*sz[0]];
			}
			else
			    dest[x+y*sz[0]] = 0;

 		}


}


function testFFT3()
{
  var nii = KViewer.viewports[0].getCurrentViewer().nii;
 

  var props = {voxSize:[nii.voxSize[0],nii.voxSize[1],1],
					  sizes:[nii.sizes[0],nii.sizes[1],1],
					  edges:nii.edges };
  var slice = createNifti(props ,  "test1" , 'float');

  var slice2 = createNifti(props ,  "test2" , 'float');

  var sz = nii.sizes;


  var tmp1 = new Float64Array(2*sz[0]*sz[1]);
  var tmp2 = new Float64Array(2*sz[0]*sz[1]);
  var slice_fft = new Float64Array(2*sz[0]*sz[1]);


  var slab = 40;
  var center = Math.floor(nii.sizes[2]/2);


  for (var x=0;x<sz[0];x++)
	  for (var y=0;y<sz[1];y++)
	  {
		  var m = 999999;
		  for (var z= center-slab; z < center+slab; z++)
		  {
  		    var idx = x+sz[0]*y+sz[0]*sz[1]*z;
  		    if (nii.data[idx] <m)
  		    	m =nii.data[idx];
		  }
		  slice.data[x+sz[0]*y] = m;
	  }

   slice.updateLimits();

   rotslice(slice.data,slice2.data,slice.sizes,0.5);
   
   slice2.updateLimits();


	var sz = slice.sizes;

   fft_2d(slice.data,sz,slice_fft,false);

   var N = 55;

   var arr = [];

   var maxy = -99999;
   var max_j ;

   for (var j=0;j <N;j++)
   { 
       rotslice(slice2.data,tmp1,sz,(N-j)/N*2*Math.PI);
       fft_2d(tmp1,sz,tmp2,false);
       for (var k = 0; k < sz[0]*sz[1];k++)
       {
       	  var a = tmp2[2*k]  /10000000000000;
       	  var b = tmp2[2*k+1]/10000000000000;
       	  tmp2[2*k]   = slice_fft[2*k]*a +  slice_fft[2*k+1]*b;
       	  tmp2[2*k+1] = slice_fft[2*k]*b -  slice_fft[2*k+1]*a;
       } 
       ifft_2d(tmp2,sz,tmp1,true);

	   var m = -9999999; 
	   var mx,my;
  	   for (var y = 0; y < sz[1];y++)
 		  for (var x = 0; x < sz[0];x++)
 		  {
			  var r = tmp1[2*(x+y*sz[0])]*tmp1[2*(x+y*sz[0])] + tmp1[2*(x+y*sz[0])+1]*tmp1[2*(x+y*sz[0])+1];
			  if (r>m)
			  {
			  	  m =r;
			  	  mx = x;
			  	  my = y;
			  }
 		  }

		if (m>maxy)
		{
			maxy=m;
			max_j=j;
		}


	    arr.push(m);

   }

   console.log(max_j/N*2*Math.PI);













  return;


  prepare(nii);
  return;


  var sz = nii.sizes;


  var tmp = new Float64Array(2*sz[0]*sz[1]*sz[2]);
  var tmp2 = new Float64Array(2*sz[0]*sz[1]*sz[2]);

  fft_3d(nii.data,sz,tmp,false)

  var x2 = Math.floor(sz[0]/2);
  var y2 = Math.floor(sz[1]/2);
  var z2 = Math.floor(sz[2]/2);

  
  for (var z = 0; z < sz[2];z++)
  for (var y = 0; y < sz[1];y++)
  for (var x = 0; x < sz[0];x++)
	 { 
	   var k = x+y*sz[0]+z*sz[0]*sz[1];
	   var xi = Math.cos(x/sz[0]*2*Math.PI);
	   var yi = Math.cos(y/sz[1]*2*Math.PI);
	   var zi = Math.cos(z/sz[2]*2*Math.PI);
	   var f = Math.pow(xi*yi*zi,4);
	   tmp[2*k] = tmp[2*k]*tmp[2*k] + tmp[2*k+1]*tmp[2*k+1];
	   tmp[2*k+1] = 0;
	 }

  ifft_3d(tmp,sz,tmp2,true)

 
  for (var z = 0; z < sz[2];z++)
  for (var y = 0; y < sz[1];y++)
  for (var x = 0; x < sz[0];x++)
	 { 
	    var i = x+y*sz[0]+z*sz[0]*sz[1];
	    var j = (x+x2)%sz[0]+((y+y2)%sz[1])*sz[0]+((z+z2)%sz[2])*sz[0]*sz[1];
	 	nii.data[i] = tmp2[2*j]/100000;

	 }





}
//setTimeout(testFFT3,1500)