<?php

// =================  session management
//ini_set('error_reporting', PHP_ERROR_REPORTING);

$page = ( "nora" . basename(dirname($_SERVER['PHP_SELF'])) );
session_name($page); // This is not good anymore for some reason.

// session timeout management: 2 stages:
// (1) cookie_lifetime defines the lifetime for the cookie send to the server.
// (2) additionaly, server performs garbage collection (gc) on cookie files on server
//     with a probability of  gc_probability / gc_divisor, garbage collector checks the cookies and deletes any who lived longer than gc_maxlifetime
//     so, gc_maxlifetime MUST BE LONGER than cookie_lifetime, otherwise, cookies are deleted randomly from that time.
//     if we want to overwrite these values, maybe must also define a save_path for the cookies, but this is not clear yet. 
//    Stick with standard path for now


//define('COOKIEPATH', '/tmp');
define('COOKIEPATH', (dirname(__FILE__) ) . "/../var/cookies/" );

ini_set("session.save_path", COOKIEPATH);  // standard is /var/lib/php5. Must be set to e.g. tmp, otherwise gc_maxlifetime does not work

//$maxlifetime = 14 * 3600 * 24; // two weeks
$maxlifetime = 52* 7 * 3600 * 24; // 52 weeks

ini_set("session.gc_maxlifetime", $maxlifetime); 
ini_set("session.cookie_lifetime", $maxlifetime);

// standards for these params can be very different for php versions. Choose  0.1%
ini_set("session.gc_probability", 100);
ini_set("session.gc_divisor", 100);


// global setting for cookie lifetime. not necessary, as will be defined below.

// define script path for the cookie ==> each domain will have its own cookie, e.g. /htocs/miracs, /DPX ...
// define also cookie lifetime here. 0 stands for forever, however this will interfere with garbage collector.

//$scriptPath =  substr ( $_SERVER['PHP_SELF'] , 0 , strrpos($_SERVER['PHP_SELF'] ,'/') ) . '/';
//session_set_cookie_params ( 10 , $scriptPath);

session_start();

$_SESSION['mem'] =  memory_get_peak_usage();



// get some info about the connected user
$_SESSION['HTTP_USER_AGENT'] =  $_SERVER['HTTP_USER_AGENT'];


if ( isset($_SERVER['HTTP_X_FORWARDED_FOR'])) {
	$_SESSION['REMOTE_ADDR'] = $_SERVER['HTTP_X_FORWARDED_FOR'];
} 
else{ 
	$_SESSION['REMOTE_ADDR'] = $_SERVER['REMOTE_ADDR'];
}


//list_active_sessions();
//die( $_SERVER['HTTP_USER_AGENT']);

function session_manager_list()
{
	// first, save the current session;
	$current_session = session_encode();
	

	$list = Array();

	try
	{
		$slist = glob(COOKIEPATH . '/sess*');
		//var_dump($slist);
		$list['list'] = Array();
		$list['header'] = Array('username', 'lastfiletime', 'lastaccess', 'filepath');
		foreach($slist as $val)
		{
			$c = file_get_contents($val);

			// clear last session entries
			foreach ($_SESSION as $key => $value) 	{ unset($_SESSION[$key]);	}

			session_decode($c);

			$obj = Array();

			//var_dump($_SESSION);
			if( isset($_SESSION['username']) )
				$obj['username'] 	 = $_SESSION['username'];
			else
				$obj['username'] 	 = "undefined";

			$obj['lastfiletime'] = date('Y m d - H:i',  filemtime($val));
			$obj['lastaccess']   = time() - filemtime($val) ;

			$obj['HTTP_USER_AGENT'] = isset($_SESSION['HTTP_USER_AGENT'])?$_SESSION['HTTP_USER_AGENT']:"";
			$obj['REMOTE_ADDR']  	= isset($_SESSION['REMOTE_ADDR'])?$_SESSION['REMOTE_ADDR']:"";

			$filename     = explode('/', $val);
			$obj['filename'] = end($filename);

			$obj['sessionid']    = session_id();

			//var_dump($obj);
			$list['list'][] = $obj;
		}
	}
	catch( Exception $e )
	{
		// restore the current session		
		foreach ($_SESSION as $key => $value){ unset($_SESSION[$key]); }
		session_decode($current_session);
		throw new Exception($e);

	}

	// restore the current session		
	foreach ($_SESSION as $key => $value){ unset($_SESSION[$key]); }
	session_decode($current_session);

	return $list;
}

function session_manager_delete($obj)
{

	$out = Array();
	$out['success'] = false;
	$out['msg'] = "";
	
	$out['msg'] .= "deleting ". $obj['filename']  ;

	//just in case ...
	if(strlen(COOKIEPATH) < 3)
	{
		$out['msg'] .= ' cookiepath too short ';
		return $out;
	}

	$slist = glob(COOKIEPATH . '/sess*');
	//var_dump($slist);
	$list['list'] = Array();
	$list['header'] = Array('username', 'lastfiletime', 'lastaccess', 'filepath');
	foreach($slist as $val)
	{
		$filename = (explode('/', $val));
		$filename = end($filename);
		if($filename == $obj['filename'])
			if( unlink($val) )
			    $out['success'] = true;
	}
	return $out;

}







// test for session expiration and cookies.
// if(isset($_GET['test']))
// {
//     session_destroy();
//     session_start();

//     $time_start = microtime(true);
//     $_SESSION['time_start'] = $time_start;
// }
// if(isset($_SESSION['time_start']))
// {
//     $_SESSION['time_diff'] = microtime(true) - $_SESSION['time_start'];
// }   
//  $meinwert = "Hallo Welt";

// setcookie("Test", $meinwert, time()+1); # 1 Stunde gültig 

// echo "cookie: " . $_COOKIE["Test"]; 
// echo "<br><br>";

// var_dump($_SESSION);
// echo "<br><br>";
// echo session_id();
// echo "<br><br>";

// var_dump( session_get_cookie_params());
// die();
    


?>